/** 
 * vpop-xmlrpc
 *
 *
 * $Id: vpop-xmlrpc.c,v 1.5 2005/04/08 15:57:45 mike Exp $
 */

/* {{{ includes */
#include <config.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <iconv.h>
#include <vpopmail/vpopmail.h>
#include <vpopmail/config.h>
#include <vpopmail/vauth.h>
#include <xmlrpc.h>
#include <xmlrpc_cgi.h>
/* }}} */

#ifndef VPOP_XMLRPC_ENCODING
#define VPOP_XMLRPC_ENCODING "ISO-8859-1"
#endif

/* {{{ MACROS */
#define RETURN_TRUE \
    return xmlrpc_build_value(env, "b", (xmlrpc_bool) 1);
#define RETURN_FALSE \
    return xmlrpc_build_value(env, "b", (xmlrpc_bool) 0);
#define RETURN_V_FAULT(rs) \
    if (handle_verror(env, (rs))) { \
        return NULL; \
    }
#define RETURN_V_FAULT_MSG(rs, msg) \
    if (rs) { \
        xmlrpc_env_set_fault(env, 0, msg); \
        VAUTH_CLOSE(); \
        return NULL;\
    }
#define RETURN_V_FAULT_MSG_FREE_1(rs, msg, tofree) \
    if (rs) { \
        xmlrpc_env_set_fault(env, 0, msg); \
        free(tofree); \
        VAUTH_CLOSE(); \
        return NULL; \
    }
#define RETURN_FAULT \
    if (env->fault_occurred) { \
        return NULL; \
    }
#define ICONV_OPEN(cd) \
    RETURN_V_FAULT_MSG((cd = iconv_open(VPOP_XMLRPC_ENCODING, "UTF-8")) == (iconv_t) -1, "Failed to init iconv")
#define ICONV_CLOSE(cd) \
    iconv_close(cd)
#define ICONV(cd, in, out) \
    { \
        size_t in_len = strlen(in); \
        size_t out_len = 2 * in_len * sizeof(int); \
        char *ptr; \
        out = calloc(1, out_len + 1); \
        ptr = out; \
        iconv(cd, &in, &in_len, &ptr, &out_len); \
    }
#ifdef HAVE_VAUTH_OPEN
#   define VAUTH_CLOSE()    vauth_close()
#   define VAUTH_OPEN(mod)  RETURN_V_FAULT_MSG(vauth_open(mod), "vauth_open() failed")
#else
#   define VAUTH_CLOSE()
#   define VAUTH_OPEN(mod)
#endif
    
/* }}} */

/* {{{ static inline int handle_verror(xmlrpc_env *, int) */
static inline int handle_verror(xmlrpc_env *env, int errno)
{
    if (errno) xmlrpc_env_set_fault(env, errno * -1 + 1000, verror(errno));
    return errno;
}
/* }}} */

/* {{{ vpop.adduser */
static xmlrpc_value *vpop_adduser(xmlrpc_env *env, xmlrpc_value *param, void *data)
{
    char *user, *domain, *password, *fullname;
    char *real_user, *real_domain, *real_password, *real_fullname;
    iconv_t cd;
    int vrs;
    
    xmlrpc_parse_value(env, param, "({s:s,s:s,s:s,s:s,*})",
            "user",     &user,
            "domain",   &domain,
            "password", &password,
            "fullname", &fullname
    );
    RETURN_FAULT;
    
    VAUTH_OPEN(1);
    
    ICONV_OPEN(cd);
    ICONV(cd, user, real_user);
    ICONV(cd, domain, real_domain);
    ICONV(cd, password, real_password);
    ICONV(cd, fullname, real_fullname);
    ICONV_CLOSE(cd);
   
    vrs = vadduser(real_user, real_domain, real_password, real_fullname, 0);
    
    free(real_user);
    free(real_domain);
    free(real_password);
    free(real_fullname);

    RETURN_V_FAULT(vrs);
    
    RETURN_TRUE;
}
/* }}} */

/* {{{ vpop.deluser */
static xmlrpc_value *vpop_deluser(xmlrpc_env *env, xmlrpc_value *param, void *data)
{
    char *user, *domain;
    char *real_user, *real_domain;
    iconv_t cd;
    int vrs;

    xmlrpc_parse_value(env, param, "({s:s,s:s,*})",
            "user",     &user,
            "domain",   &domain
    );
    RETURN_FAULT;
    
    VAUTH_OPEN(1);
    
    ICONV_OPEN(cd);
    ICONV(cd, user, real_user);
    ICONV(cd, domain, real_domain);
    ICONV_CLOSE(cd);
    
    vrs = vdeluser(real_user, real_domain);

    VAUTH_CLOSE();
    
    free(real_user);
    free(real_domain);
    
    RETURN_V_FAULT(vrs);
    
    RETURN_TRUE;
}
/* }}} */

/* {{{ vpop.passwd */
static xmlrpc_value *vpop_passwd(xmlrpc_env *env, xmlrpc_value *param, void *data)
{
    char *user, *domain, *password;
    char *real_user, *real_domain, *real_password;
    iconv_t cd;
    int vrs;

    xmlrpc_parse_value(env, param, "({s:s,s:s,s:s,*})",
            "user",     &user,
            "domain",   &domain,
            "password", &password
    );
    RETURN_FAULT;
    
    VAUTH_OPEN();
    
    ICONV_OPEN(cd);
    ICONV(cd, user, real_user);
    ICONV(cd, domain, real_domain);
    ICONV(cd, password, real_password);
    ICONV_CLOSE(cd);
    
    vrs = vpasswd(real_user, real_domain, real_password, 0);

    free(real_user);
    free(real_domain);
    free(real_password);
    
    RETURN_V_FAULT(vrs);
    
    RETURN_TRUE;
}
/* }}} */

/* {{{ vpop.setquota */
static xmlrpc_value *vpop_setquota(xmlrpc_env *env, xmlrpc_value *param, void *data)
{
    char *user, *domain, *quota;
    char *real_user, *real_domain, *real_quota;
    iconv_t cd;
    int vrs;

    xmlrpc_parse_value(env, param, "({s:s,s:s,s:s,*})",
            "user",     &user,
            "domain",   &domain,
            "quota",    &quota
    );
    RETURN_FAULT;
    
    VAUTH_OPEN(1);
    
    ICONV_OPEN(cd);
    ICONV(cd, user, real_user);
    ICONV(cd, domain, real_domain);
    ICONV(cd, quota, real_quota);
    ICONV_CLOSE(cd);
    
    vrs = vsetuserquota(real_user, real_domain, real_quota);

    free(real_user);
    free(real_domain);
    free(real_quota);

    RETURN_V_FAULT(vrs);
    
    RETURN_TRUE;
}
/* }}} */

/* {{{ vpop.auth */
static xmlrpc_value *vpop_auth(xmlrpc_env *env, xmlrpc_value *param, void *data)
{
    char *user, *domain, *password;
    char *real_user, *real_domain, *real_password;
    iconv_t cd;
    int vrs;

    xmlrpc_parse_value(env, param, "({s:s,s:s,s:s,*})",
            "user",     &user,
            "domain",   &domain,
            "password", &password
    );
    RETURN_FAULT;

    VAUTH_OPEN(0);
    
    ICONV_OPEN(cd);
    ICONV(cd, user, real_user);
    ICONV(cd, domain, real_domain);
    ICONV(cd, password, real_password);
    ICONV_CLOSE(cd);
    
    vrs = vauth_crypt(real_user, real_domain, real_password, vauth_getpw(real_user, real_domain));

    VAUTH_CLOSE();
    
    free(real_user);
    free(real_domain);
    free(real_password);
    
    if (vrs == 0) {
        RETURN_TRUE;
    } else {
        RETURN_FALSE;
    }
}
/* }}} */

/* {{{ vpop.adddomain * /
static xmlrpc_value *vpop_adddomain(xmlrpc_env *env, xmlrpc_value *param, void *data)
{
    char *domain, *password;

    xmlrpc_parse_value(env, param, "({s:s,s:s,*})",
            "domain",   &domain,
            "password", &password
    );
    
    RETURN_FAULT;
    RETURN_V_FAULT(vadddomain(domain, VPOPMAILDIR, VPOPMAILUID, VPOPMAILGID));
    RETURN_V_FAULT(vadduser("postmaster", domain, password, "Postmaster", 0));
    RETURN_TRUE;
}
/ * }}} */

/* {{{ vpop.deldomain * /
static xmlrpc_value *vpop_deldomain(xmlrpc_env *env, xmlrpc_value *param, void *data)
{
    char *domain;

    xmlrpc_parse_value(env, param, "({s:s,*})",
            "domain",   &domain
    );

    RETURN_FAULT;
    RETURN_V_FAULT(vdeldomain(domain));
    RETURN_TRUE;
}
/ * }}} */

/* {{{ vpop.listusers */
static xmlrpc_value *vpop_listusers(xmlrpc_env *env, xmlrpc_value *param, void *data)
{
    char *domain, *in_domain;
    int first = 1;
    uid_t pw_uid;
    gid_t pw_gid;
    struct vqpasswd *mypw;
    xmlrpc_value *users;
    iconv_t cd;

    xmlrpc_parse_value(env, param, "({s:s,*})", "domain", &in_domain);
    RETURN_FAULT;

    VAUTH_OPEN(0);

    ICONV_OPEN(cd);
    ICONV(cd, in_domain, domain);
    ICONV_CLOSE(cd);
    
    vget_assign(domain, NULL, 0, &pw_uid, &pw_gid);
    
    RETURN_V_FAULT_MSG_FREE_1(setgid(pw_gid), "setgid() failed", domain);
    RETURN_V_FAULT_MSG_FREE_1(setuid(pw_uid), "setuid() failed", domain);

    users = xmlrpc_build_value(env, "{}");
    
    while ((mypw = vauth_getall(domain, first, 1))) {
        first = 0;
        xmlrpc_value *user = xmlrpc_build_value(env, "{}");
        xmlrpc_value *username = xmlrpc_build_value(env, "s", mypw->pw_name);

        xmlrpc_struct_set_value_v(env, user, 
                xmlrpc_build_value(env, "s", "name"),
                xmlrpc_build_value(env, "s", mypw->pw_name)
        );
        xmlrpc_struct_set_value_v(env, user, 
                xmlrpc_build_value(env, "s", "passwd"),
                xmlrpc_build_value(env, "s", mypw->pw_passwd)
        );
        xmlrpc_struct_set_value_v(env, user, 
                xmlrpc_build_value(env, "s", "clear passwd"),
                xmlrpc_build_value(env, "s", mypw->pw_clear_passwd)
        );
        xmlrpc_struct_set_value_v(env, user, 
                xmlrpc_build_value(env, "s", "uid"),
                xmlrpc_build_value(env, "i", mypw->pw_uid)
        );
        xmlrpc_struct_set_value_v(env, user, 
                xmlrpc_build_value(env, "s", "gid"),
                xmlrpc_build_value(env, "i", mypw->pw_gid)
        );
        xmlrpc_struct_set_value_v(env, user, 
                xmlrpc_build_value(env, "s", "flags"),
                xmlrpc_build_value(env, "i", mypw->pw_flags)
        );
        xmlrpc_struct_set_value_v(env, user, 
                xmlrpc_build_value(env, "s", "quota"),
                xmlrpc_build_value(env, "s", mypw->pw_shell)
        );
        if (strcmp(mypw->pw_shell, "NOQUOTA")) {
            char quota[5], maildir[255];
            snprintf(maildir, sizeof(maildir), "%s/Maildir", mypw->pw_dir);
            snprintf(quota, sizeof(quota), "%d%%", vmaildir_readquota(maildir, format_maildirquota(mypw->pw_shell)));
            xmlrpc_struct_set_value_v(env, user, 
                    xmlrpc_build_value(env, "s", "usage"),
                    xmlrpc_build_value(env, "s", quota)
            );
        } else {
            xmlrpc_struct_set_value_v(env, user, 
                    xmlrpc_build_value(env, "s", "usage"),
                    xmlrpc_build_value(env, "s", mypw->pw_shell)
            );
        }
#ifdef ENABLE_AUTH_LOGGING
        {
            char *authip;
            time_t authtime;
            authtime = vget_lastauth(mypw, domain);
            authip   = vget_lastauthip(mypw, domain);
            if ((!authtime) || (!authip) || (!strcmp(authip, "0.0.0.0"))) {
                xmlrpc_struct_set_value_v(env, user, 
                        xmlrpc_build_value(env, "s", "last auth"),
                        xmlrpc_build_value(env, "s", "Never logged in")
                );
            } else {
                xmlrpc_struct_set_value_v(env, user,
                        xmlrpc_build_value(env, "s", "last auth"),
                        xmlrpc_build_value(env, "s", asctime(localtime(&authtime)))
                );
            }
        }
#else
        xmlrpc_struct_set_value_v(env, user, 
                xmlrpc_build_value(env, "s", "last auth"),
                xmlrpc_build_value(env, "s", "Not available")
        );
#endif
        
        xmlrpc_struct_set_value_v(env, users, username, user);
    }
    
    free(domain);
    VAUTH_CLOSE();
    RETURN_FAULT;
    
    return users;
}
/* }}} */
extern vdir_type vdir;
/* {{{ vpop.listdomains */
static xmlrpc_value *vpop_listdomains(xmlrpc_env *env, xmlrpc_value *param, void *data)
{
    char start[1] = {0}, path[255], line[1024], *ptr;
    xmlrpc_value *domains;
    FILE *fs = NULL;

    snprintf(path, sizeof(path), "%s/users/assign", QMAILDIR);
    RETURN_V_FAULT_MSG(!(fs = fopen(path, "r")), "Could not open assign.");
            
    domains = xmlrpc_build_value(env, "{}");
    /* based on Toms get_domain_entries() */
    while (fgets(line, sizeof(line), fs)) {
        char *dom, *real;
        int uid, gid;
        
        if (*line != '+') {
            continue;
        }

        if ((!(dom = strtok(line + 1, ":"))) || (!strchr(dom, '.'))) {
            continue;
        }
        if (!(real = strtok(NULL, ":"))) {
            continue;
        } else {
            *(real - 2) = 0;
        }
        if (!(ptr = strtok(NULL, ":"))) {
            continue;
        } else {
            uid = atoi(ptr);
        }
        if (!(ptr = strtok(NULL, ":"))) {
            continue;
        } else {
            gid = atoi(ptr);
        }
        
        xmlrpc_value *domain = xmlrpc_build_value(env, "{}");
        xmlrpc_value *domainname = xmlrpc_build_value(env, "s", dom);
        
        xmlrpc_struct_set_value_v(env, domain,
                xmlrpc_build_value(env, "s", "domain"),
                xmlrpc_build_value(env, "s", dom)
        );
        xmlrpc_struct_set_value_v(env, domain,
                xmlrpc_build_value(env, "s", "realdomain"),
                xmlrpc_build_value(env, "s", real)
        );
        xmlrpc_struct_set_value_v(env, domain,
                xmlrpc_build_value(env, "s", "alias"),
                xmlrpc_build_value(env, "b", strcmp(dom, real))
        );
        xmlrpc_struct_set_value_v(env, domain,
                xmlrpc_build_value(env, "s", "uid"),
                xmlrpc_build_value(env, "i", uid)
        );
        xmlrpc_struct_set_value_v(env, domain,
                xmlrpc_build_value(env, "s", "gid"),
                xmlrpc_build_value(env, "i", gid)
        );
        
        open_big_dir(real, uid, gid);
        xmlrpc_struct_set_value_v(env, domain,
                xmlrpc_build_value(env, "s", "users"),
                xmlrpc_build_value(env, "i", vdir.cur_users)
        );
        close_big_dir(real, uid, gid);

        xmlrpc_struct_set_value_v(env, domains, domainname, domain);
    }

    fclose(fs);
    
    return domains;
}
/* }}} */

/* {{{ main */
int main(int argc, char **argv)
{
    /* Init CGI lib */
    xmlrpc_cgi_init(XMLRPC_CGI_NO_FLAGS);

    /* Install methods */
    xmlrpc_cgi_add_method_w_doc("vpop.adduser", &vpop_adduser, NULL,
            "b:S", "Add a vpopmail user to a domain with password and fullname.");
    xmlrpc_cgi_add_method_w_doc("vpop.deluser", &vpop_deluser, NULL,
            "b:S", "Delete a user from a domain.");
    xmlrpc_cgi_add_method_w_doc("vpop.passwd", &vpop_passwd, NULL,
            "b:S", "Change the password of a user in a domain.");
    xmlrpc_cgi_add_method_w_doc("vpop.setquota", &vpop_setquota, NULL,
            "b:S", "Set the quota of a user in a domain.");
    xmlrpc_cgi_add_method_w_doc("vpop.auth", &vpop_auth, NULL,
            "b:S", "Authenticate a user in a domain with the password.");
    /*
    xmlrpc_cgi_add_method_w_doc("vpop.adddomain", &vpop_adddomain, NULL,
            "b:S", "Add a domain with a password for the postmaster account.");
    xmlrpc_cgi_add_method_w_doc("vpop.deldomain", &vpop_deldomain, NULL,
            "b:S", "Delete an entire domain.");
    */
    xmlrpc_cgi_add_method_w_doc("vpop.listusers", &vpop_listusers, NULL,
            "S:S", "List users of a domain.");
    xmlrpc_cgi_add_method_w_doc("vpop.listdomains", &vpop_listdomains, NULL,
            "S:", "List domains.");
    
    /* Process */
    xmlrpc_cgi_process_call();

    /* Cleanup */
    xmlrpc_cgi_cleanup();
    return 0;
}
/* }}} */
